package edu.berkeley.nlp.util;

import java.util.Collection;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Callable;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.Future;

/**
 * This class manages a pool of task workers.
 */
public class ThTaskWorkerPool {
	private int mThNum;
	private ExecutorService mThTaskWorkerPool;

	/**
	 * Constructor
	 *
	 */
	public ThTaskWorkerPool() {
		setThNum(0);
	}

	/**
	 * Constructor
	 *
	 * @param thNum
	 *         specify the number of threads
	 */
	public ThTaskWorkerPool(int thNum) {
		setThNum(thNum);
	}

	/**
	 * Close worker pool
	 */
	public void close() {
		if(mThTaskWorkerPool != null) {
			mThTaskWorkerPool.shutdown();
			try {
				// Wait a while for existing tasks to terminate
				if (!mThTaskWorkerPool.awaitTermination(60, TimeUnit.SECONDS)) {
					mThTaskWorkerPool.shutdownNow(); // Cancel currently executing tasks
					// Wait a while for tasks to respond to being cancelled
					if (!mThTaskWorkerPool.awaitTermination(60, TimeUnit.SECONDS))
						System.err.println("Pool did not terminate");
				}
			} catch(Exception e) {
				mThTaskWorkerPool.shutdownNow();
			}
		}
		mThTaskWorkerPool = null;
	}

	/**
	 * Set thread number to initialize.
	 *
	 * @param thNum
	 *         specify the number of threads
	 */
	public void setThNum(int thNum) {
		close();
		mThNum = thNum;
		final int thNumDefault = Runtime.getRuntime().availableProcessors();
		if(mThNum < 0) {
			mThNum = thNumDefault;
		}
		mThTaskWorkerPool = (mThNum > 0) ? Executors.newFixedThreadPool(mThNum) : null;
	}

	/**
	 * Get the number of threads realized.
	 *
	 * @return the number of threads realized.
	 */
	public int getThNum() {
		return mThNum;
	}

	/**
	 * Get a task worker pool as ExecutorService
	 *
	 * @return a task worker pool as ExecutorService
	 */
	public ExecutorService getThTaskWorkerPool() {
		return mThTaskWorkerPool;
	}

	/**
	 * Get an object to synchronize
	 *
	 * @return an object to synchronize
	 */
	public Object getSync() {
		return this;
	}

	/**
	 * Invoke all of the task list.
	 *
	 * @param taskList
	 *         specify the list of callable task.
	 */
	public <T> List<Future<T>> invokeAll(Collection<? extends Callable<T>> taskList)
		throws InterruptedException {
		return mThTaskWorkerPool.invokeAll(taskList);
	}

	/**
	 * Execute all of the task list.
	 *
	 * @param taskList
	 *         specify the list of callable task.
	 */
	public <T> List<Future<T>> executeTaskAll(Collection<? extends Callable<T>> taskList) {
		List<Future<T>> resList = null;
		try {
			resList = invokeAll(taskList);
		} catch(InterruptedException e) {
			System.err.println("Interrupted:");
			e.printStackTrace();
		} finally {
			return resList;
		}
	}

}

